from data_generation_tool.constraints.column_constraints import Allowed, Forbidden, MinSize, MaxSize, SizeInterval
from data_generation_tool.errors import UnsatisfiableConstraints
from data_generation_tool.generators import ColumnGenerator
from data_generation_tool.requests import ColumnGenerationRequest


class QualityGenerator(ColumnGenerator):
    """
    Generator for quality columns.
    """

    def generate(self, column: ColumnGenerationRequest, count: int) -> list:
        """
        Generates a list of qualities based on the constraints provided in the column request.

        Parameters
        ----------
        column : ColumnGenerationRequest
            The request for generating the given column : contains the constraints
        count : int
            The number of values to return

        Returns
        -------
        list
            The data generated

        Raises
        -------
            ValueError
                If the expected type in the column does not match the type of values generated by the generator.

            UnsatisfiableConstraints
                If the given set of constraints are not satisfiable.
        """

        allowed_qualities = _qualities
        forbidden_qualities = []
        min_size = None
        max_size = None

        for constraint in column.constraints:
            if isinstance(constraint, Allowed):
                allowed_qualities = [quality for quality in allowed_qualities if quality in constraint.value]
            elif isinstance(constraint, Forbidden):
                forbidden_qualities = constraint.value
            elif isinstance(constraint, MinSize):
                min_size = constraint.value
            elif isinstance(constraint, MaxSize):
                max_size = constraint.value
            elif isinstance(constraint, SizeInterval):
                min_size = constraint.lower_bound
                max_size = constraint.upper_bound

        allowed_qualities = [quality for quality in allowed_qualities if quality not in forbidden_qualities]

        if min_size is not None:
            allowed_qualities = [quality for quality in allowed_qualities if len(quality) >= min_size]

        if max_size is not None:
            allowed_qualities = [quality for quality in allowed_qualities if len(quality) <= max_size]

        if not allowed_qualities:
            raise UnsatisfiableConstraints("The constraints are unsatisfiable.")

        return self.post_processing(self.rng.choice(allowed_qualities, count), column.constraints)


_qualities = [
    "Excellent",
    "Good",
    "Fair",
    "Average",
    "Poor",
    "High",
    "Low",
    "Superior",
    "Inferior",
    "Premium",
    "Luxury",
    "Economic",
    "Standard",
    "Top-tier",
    "Mid-range"
]
